from typing import *
from ...queryable import Queryable

T = TypeVar("T")

class partition_by_count:
    def __init__(self, count: int):
        self.count = count

    def _make(self, query):
        buffer = []
        for e in query:
            buffer.append(e)
            if len(buffer) == self.count:
                yield buffer
                buffer = []
        if len(buffer) > 0:
            yield buffer

    def __call__(self, query):
        return Queryable(self._make(query))



class partition_by_selector:
    def __init__(self, selector: Callable[[T],Any]):
        self.selector = selector

    def _make(self, query):
        buffer = []
        first_time = True
        old_field = None
        for e in query:
            if first_time:
                old_field = self.selector(e)
                buffer.append(e)
                first_time = False
                continue
            new_field = self.selector(e)
            if new_field!=old_field:
                yield buffer
                buffer = []
            buffer.append(e)
            old_field=new_field
        if len(buffer)>0:
            yield buffer

    def __call__(self, query):
        return Queryable(self._make(query))




from .arch import PushQueryElement, PushQueryElement
from collections import OrderedDict
from typing import *


class ObservableBase(PushQueryElement):
    def on_exit(factory, instance, exc_type, exc_val, exc_tb):
        for key, value in instance.subscribers.items():
            value.__exit__(exc_type, exc_val, exc_tb)


class SplitPipelines(ObservableBase):
    def __init__(self, *args: PushQueryElement, **kwargs: PushQueryElement):
        key_values = []
        for e in args:
            key = type(e).__name__.lower()
            key_values.append((key,e))
        for key, value in kwargs.items():
            key_values.append((key,value))
        self.pipelines = OrderedDict()
        for key, value in key_values:
            final_key = key
            for i in range(0,100):
                final_key = key+('' if i==0 else '_'+str(i))
                if final_key not in self.pipelines:
                    break
            self.pipelines[final_key]=value


    def on_enter(factory,instance):
        instance.subscribers = OrderedDict()
        for key, value in factory.pipelines.items():
            inst = value.instance()
            inst.__enter__()
            instance.subscribers[key] = inst

    def on_process(factory, instance, element):
        for value in instance.subscribers.values():
            value.process(element)

    def on_report(factory, instance):
        return {key: value.report() for key, value in instance.subscribers.items()}


class ObservableAbstract(ObservableBase):
    def __init__(self):
        self.subscribers = OrderedDict()  # type: Dict[Any,PushQueryElement]

    def subscribe(self, pqe: PushQueryElement, name):
        self.subscribers[name] = pqe

    def on_enter(factory, instance):
        instance.subscribers = OrderedDict()
        for key, value in factory.subscribers.items():
            inst = value.instance()
            inst.__enter__()
            instance.subscribers[key] = inst

    def push_element(factory, instance, element):
        for value in instance.subscribers.values():
            value.process(element)



class SelectPQE(ObservableAbstract):
    def __init__(self, selector: Callable):
        super(SelectPQE, self).__init__()
        self.selector = selector

    def on_process(factory, instance, element):
        factory.push_element(instance, factory.selector(element))

    def on_report(factory, instance):
        for value in instance.subscribers.values():
            return value.report()



class SelectManyPQE(ObservableAbstract):
    def __init__(self, selector: Callable):
        super(SelectManyPQE, self).__init__()
        self.selector = selector

    def on_process(factory, instance, element):
        for smaller_element in factory.selector(element):
            factory.push_element(instance, smaller_element)

    def on_report(factory, instance):
        for value in instance.subscribers.values():
            return value.report()



class WherePQE(ObservableAbstract):
    def __init__(self, filter: Callable):
        super(WherePQE, self).__init__()
        self.filter = filter

    def on_process(factory, instance, element):
        if factory.filter(element):
            factory.push_element(instance, element)

    def on_report(factory, instance):
        for value in instance.subscribers.values():
            return value.report()



class Dispatch(ObservableBase):
    def __init__(self):
        self.continuation = None  # type: Optional[PushQueryElement]

    def subscribe(self, pqe, name):
        if self.continuation is None:
            self.continuation = pqe
        else: #pragma: no cover
            raise ValueError('There is already one subscriber, and Dispatch does not accept more')


    def on_enter(factory, instance):
        instance.subscribers = OrderedDict()

    def push_to_bucket(factory, instance, element, bucket):
        if factory.continuation is None: #pragma: no cover
            raise ValueError('Cannot route: the continuation is not set')
        if bucket not in instance.subscribers:
            instance.subscribers[bucket] = factory.continuation.instance()
            instance.subscribers[bucket].__enter__()
        instance.subscribers[bucket].process(element)


    def on_report(factory, instance):
        return {key: value.report() for key, value in instance.subscribers.items()}



class SplitByGroup(Dispatch):
    def __init__(self, group_selector, with_total=None):
        super(SplitByGroup, self).__init__()
        self.group_selector = group_selector
        self.with_total = with_total

    def on_process(factory, instance, element):
        bucket = factory.group_selector(element)
        factory.push_to_bucket(instance, element, bucket)
        if factory.with_total is not None:
            factory.push_to_bucket(instance, element, factory.with_total)


class SplitByDictionary(Dispatch):
    def __init__(self):
        super(SplitByDictionary, self).__init__()

    def on_process(factory, instance, element):
        if not isinstance(element, dict):
            raise ValueError('SplitByDictionary expects disctionaries!')
        for key, value in element.items():
            factory.push_to_bucket(instance, value, key)



from typing import *
from collections import OrderedDict
from ...queryable import Queryable
from ...query import Query

def strjoin(separator: str):
    return lambda en: separator.join([str(s) for s in en])


class pairwise:
    def __init__(self):
        pass

    def _make(self, en):
        old = None
        firstTime = True
        for e in en:
            if not firstTime:
                yield (old, e)
            old = e
            if firstTime:
                firstTime = False

    def __call__(self, en):
        return Queryable(self._make(en))


class count_by:
    def __init__(self, selector):
        self.selector = selector

    def __call__(self, en):
        result = OrderedDict()
        for e in en:
            gr = self.selector(e)
            if gr not in result:
                result[gr] = 0
            result[gr] += 1
        return Query.dict(result)



from collections import OrderedDict
import json

_INITIALIZED_KEY = '_Obj__initialized'
_EXCEPTIONS = {_INITIALIZED_KEY }

class Obj(OrderedDict):
    def __init__(self, **kwargs):
        super(Obj, self).__init__(**kwargs)
        self.__initialized = True

    def __setattr__(self, key, value):
        if _INITIALIZED_KEY not in self.__dict__:
            self.__dict__[key] = value
        else:
            if key in _EXCEPTIONS:
                raise ValueError('{0} attribute cannot be set, it is a reserved field'.format(key))
            self[key] = value

    def update(self, **kwargs):
        for key, value in kwargs.items():
            self[key]=value
        return self

    def remove(self, *args):
        for key in args:
            del self[key]
        return self

    def __repr__(self):
        return dict.__repr__(self)

    def __getattr__(self, item):
        if _INITIALIZED_KEY not in self.__dict__: #pragma: no cover
            return self.__dict__[item]
        else:
            if item in _EXCEPTIONS: #pragma: no cover
                return self.__dict__[item]
            else:
                try:
                    return self[item]
                except:
                    raise AttributeError(item)


    @staticmethod
    def _make_pretty(obj):
        if isinstance(obj,dict):
            return Obj( **{key:Obj._make_pretty(obj[key]) for key in obj})
        elif isinstance(obj,list):
            return [Obj._make_pretty(o) for o in obj]
        else:
            return obj

    @staticmethod
    def create(d):
        return Obj._make_pretty(d)


from typing import *
import copy

class _NoSpecialCase:
    pass

class AbstractPushQueryElement:
    def instance(self): #pragma: no cover
        raise NotImplementedError()

    def subscribe(self, pqe, name):
        raise ValueError('This Element is not subscribable')

    def __call__(self, enumerable: Iterable):
        with self.instance() as agg:
            for element in enumerable:
                process_result = agg.process(element)
                if isinstance(process_result,StopIteration):
                    break
            return agg.report()


class AbstractPushQueryElementInstance:
    def __enter__(self): #pragma: no cover
        raise NotImplementedError()

    def process(self, element): #pragma: no cover
        pass

    def report(self): #pragma: no cover
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class _ArithmeticPQEI(AbstractPushQueryElementInstance):
    def __init__(self,
                 enter_lambda,
                 process_lambda,
                 report_lambda,
                 throw_if_empty,
                 value_if_empty,
                 rettype,
                 stop_iteration_lambda
                 ):
        self._enter_lambda = enter_lambda
        self._process_lambda = process_lambda
        self._report_lambda = report_lambda
        self._throw_if_empty = throw_if_empty
        self._value_if_empty = value_if_empty
        self._rettype = rettype
        self._stop_iteration_lambda = stop_iteration_lambda

    def __enter__(self):
        self._count = 0
        self._state = self._enter_lambda()
        return self

    def process(self, element):
        self._count+=1
        self._state = self._process_lambda(self._state, element)
        ret_status = self._stop_iteration_lambda(self._state)
        return ret_status

    def report(self):
        if self._count == 0:
            if self._throw_if_empty:
                raise ValueError('The sequence was empty')
            elif not isinstance(self._value_if_empty,_NoSpecialCase):
                result = self._value_if_empty
            else:
                result = self._report_lambda(self._state)
        else:
            result = self._report_lambda(self._state)
        if self._rettype is not None:
            result = self._rettype(result)
        return result



class ArithmeticPQE(AbstractPushQueryElement):
    def __init__(self,
                 enter_lambda,
                 process_lambda,
                 report_lambda,
                 throw_if_empty,
                 value_if_empty = _NoSpecialCase(),
                 stop_iterations_lambda = lambda z: None):
        self._enter_lambda = enter_lambda
        self._process_lambda = process_lambda
        self._report_lambda = report_lambda
        self._throw_if_empty = throw_if_empty
        self._rettype = None
        self._value_if_empty = value_if_empty
        self._stop_iterations_lambda = stop_iterations_lambda

    def instance(self):
        return _ArithmeticPQEI(
            self._enter_lambda,
            self._process_lambda,
            self._report_lambda,
            self._throw_if_empty,
            self._value_if_empty,
            self._rettype,
            self._stop_iterations_lambda
        )

    def throw_if_empty(self):
        result = copy.deepcopy(self)
        result._throw_if_empty = True
        return result

    def default_if_empty(self, default_value = None):
        result = copy.deepcopy(self)
        result._throw_if_empty = False
        result._value_if_empty = default_value
        return result

    def astype(self, rettype):
        result = copy.deepcopy(self)
        result._rettype = rettype
        return result


class _PushQueryElementInstance(AbstractPushQueryElementInstance):
    def __init__(self, enter_function, process_function, report_function, exit_function):
        self.enter_function = enter_function
        self.process_function = process_function
        self.report_function = report_function
        self.exit_function = exit_function

    def __enter__(self):
        self.enter_function(self)
        return self

    def process(self, element):
        return self.process_function(self, element)

    def report(self):
        return self.report_function(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.exit_function(self,exc_type, exc_val, exc_tb)


class PushQueryElement(AbstractPushQueryElement):
    def instance(self):
        return _PushQueryElementInstance(
            self.on_enter,
            self.on_process,
            self.on_report,
            self.on_exit
        )

    def on_enter(factory,instance): #pragma: no cover
        raise NotImplementedError()

    def on_process(factory, instance, element): #pragma: no cover
        raise NotImplementedError()

    def on_report(factory, instance): #pragma: no cover
        raise NotImplementedError()

    def on_exit(factory, instance,exc_type, exc_val, exc_tb):
        pass






from typing import *
from .._misc import *
import itertools

def grid_iter(keys,lists):
    for config in itertools.product(*lists):
        yield Obj(**{key: value for key, value in zip(keys, config)})


def grid_dict(dict):
    keys = list(dict)
    lists = [dict[key] for key in keys]
    length = 1
    for l in lists:
        length *= len(l)
    return grid_iter(keys, lists), length



def grid_args(args):
    length = 1
    for l in args:
        length*=len(l)
    return itertools.product(*args),length



def triangle_iter(items,with_diagonal):
    for index1,item1 in enumerate(items):
        begin = index1
        if not with_diagonal:
            begin+=1
        for index2 in range(begin,len(items)):
            item2=items[index2]
            yield (item1,item2)

T = TypeVar('T')

from itertools import combinations

def powerset_iter(iterable):
    xs = list(iterable)
    for i in range(len(xs)+1):
        for p in combinations(xs,i):
            yield p


















from .._push_queries.arch import ArithmeticPQE
import math

class Sum(ArithmeticPQE):
    def __init__(self):
        super(Sum, self).__init__(
            lambda: 0,
            lambda state,element: state+element,
            lambda state: state,
            True,
            None
        )

class Count(ArithmeticPQE):
    def __init__(self):
        super(Count, self).__init__(
            lambda: 0,
            lambda state,element: state+1,
            lambda state:state,
            False
        )

class Mean(ArithmeticPQE):
    def __init__(self):
        super(Mean, self).__init__(
            lambda: (0,0),
            lambda state,element: (state[0]+element,state[1]+1),
            lambda state: state[0]/state[1],
            False,
            None
        )

class Std(ArithmeticPQE):
    @staticmethod
    def _process(state, element):
        mean, count, M2 = state
        count+=1
        delta = element - mean
        mean += delta / count
        delta2 = element - mean
        M2 += delta * delta2
        return (mean,count,M2)

    @staticmethod
    def _exit(state):
        mean, count, M2= state
        if count < 2:
            return None
        return math.sqrt(M2 / count)

    def __init__(self):
        super(Std, self).__init__(
            lambda: (0,0,0),
            Std._process,
            Std._exit,
            False
        )

class First(ArithmeticPQE):
    def __init__(self):
        super(First, self).__init__(
            lambda: (True, None),
            lambda state, element: (False,element) if state[0] else (False,state[1]),
            lambda state: state[1],
            True,
            None,
            lambda state: StopIteration() if not state[0] else None
        )

class Last(ArithmeticPQE):
    def __init__(self):
        super(Last, self).__init__(
            lambda: None,
            lambda state, element: element,
            lambda state: state,
            True
        )

class Single(ArithmeticPQE):
    @staticmethod
    def _process(state,element):
        if not state[0]:
            raise ValueError('Sequence contains more than one element')
        return (False,element)

    def __init__(self):
        super(Single, self).__init__(
            lambda: (True,None),
            Single._process,
            lambda state: state[1],
            True)


class Max(ArithmeticPQE):
    def __init__(self, value_selector = lambda z:z):
        super(Max, self).__init__(
            lambda: (True,None),
            lambda state, element: (False,value_selector(element)) if state[0] else (False,max(state[1],value_selector(element))),
            lambda state: state[1],
            True
        )

class Min(ArithmeticPQE):
    def __init__(self, value_selector = lambda z:z):
        super(Min, self).__init__(
            lambda: (True,None),
            lambda state, element: (False,value_selector(element)) if state[0] else (False,min(state[1],value_selector(element))),
            lambda state: state[1],
            True
        )

class _ProcessorArgMinMax:
    def __init__(self,value_selector,is_argmin):
        self.is_argmin = is_argmin
        self.value_selector = value_selector

    def _process(self, state,element):
        first_time, cur_value, cur_element = state
        value = self.value_selector(element)
        if first_time:
            return (False, value, element)
        elif (value < cur_value and self.is_argmin) or (value > cur_value and not self.is_argmin):
            return (False, value, element)
        else:
            return (False, cur_value, cur_element)


class ArgMin(ArithmeticPQE):
    def __init__(self, value_selector=lambda z: z):
        pr = _ProcessorArgMinMax(value_selector,True)
        super(ArgMin, self).__init__(
            lambda: (True, None, None),
            pr._process,
            lambda state: state[2],
            True
        )


class ArgMax(ArithmeticPQE):
    def __init__(self, value_selector=lambda z: z):
        pr = _ProcessorArgMinMax(value_selector,False)
        super(ArgMax, self).__init__(
            lambda: (True, None, None),
            pr._process,
            lambda state: state[2],
            True
        )

def _all_any_value_selectors(value):
    if isinstance(value,bool):
        return value
    return True

class All(ArithmeticPQE):
    def __init__(self, value_selector=None):
        if value_selector is None:
            value_selector = _all_any_value_selectors
        super(All, self).__init__(
            lambda: True,
            lambda state, element: state and value_selector(element),
            lambda state: state,
            False
        )

class Any(ArithmeticPQE):
    def __init__(self, value_selector=None):
        if value_selector is None:
            value_selector = _all_any_value_selectors
        super(Any, self).__init__(
            lambda: False,
            lambda state, element: state or value_selector(element),
            lambda state: state,
            False
        )





from typing import Optional, Callable
from .._push_queries.arch import PushQueryElement
from .._misc import KeyValuePair

class ToList(PushQueryElement):
    def on_enter(factory,instance):
        instance._list = []

    def on_process(factory, instance, element):
        instance._list.append(element)

    def on_report(factory, instance):
        return instance._list


class ToTuple(PushQueryElement):
    def on_enter(factory, instance):
        instance._list = []

    def on_process(factory, instance, element):
        instance._list.append(element)

    def on_report(factory, instance):
        return tuple(instance._list)


class ToSet(PushQueryElement):
    def on_enter(factory,instance):
        instance._set = set()

    def on_process(factory, instance, element):
        if element in instance._set:
            raise ValueError('Duplicating element {0}'.format(element))
        instance._set.add(element)

    def on_report(factory, instance):
        return instance._set

class ToDictionary(PushQueryElement):
    def __init__(self, key_selector: Optional[Callable] = None, value_selector: Optional[Callable] = None):
        if (key_selector is None) != (value_selector is None):
            raise ValueError('KeySelector and ValueSelector must be defined both or none of them')
        self.key_selector = key_selector
        self.value_selector = value_selector

    def on_enter(factory,instance):
        instance._dict = {}
        instance._first_time = True
        instance._await_key_value_pair = None
        instance.key_selector = factory.key_selector
        instance.value_selector = factory.value_selector

    def on_process(factory, instance, element):
        if instance._first_time:
            if isinstance(element,KeyValuePair):
                instance._await_key_value_pair = True
                if instance.key_selector is None:
                    instance.key_selector = lambda z: z.key
                if instance.value_selector is None:
                    instance.value_selector = lambda z: z.value
            else:
                instance._await_key_value_pair = False
                if instance.key_selector is None or instance.value_selector is None:
                    raise ValueError('KeySelector/ValueSelector are not specified, but sequence does not consist of key/value pairs')
            instance._first_time = False
        else:
            if instance._await_key_value_pair != isinstance(element,KeyValuePair):
                raise ValueError("The sequence is a mixture of KeyValue pair and other types. This is not allowed")

        key = instance.key_selector(element)
        if key in instance._dict:
            raise ValueError('Duplicating value for key {0}'.format(key))
        instance._dict[key]=instance.value_selector(element)

    def on_report(factory, instance):
        return instance._dict




from typing import *
from .._misc import KeyValuePair
from pathlib import Path
import gzip, pickle
import os
import zipfile



def text_file(filename, **kwargs):
    with open(filename, 'r', **kwargs) as file:
        for line in file:
            if line.endswith('\n'):
                line = line[:-1]
            yield line

def zip_text_file(filename, encoding):
    with gzip.open(filename, 'rb') as f:
        for line in f:
            if line.endswith(b'\n'):
                line = line[:-1]
            yield line.decode(encoding)


def pickle_file(filename, file_obj_factory):
    with file_obj_factory(filename) as file:
        while file.read(1):
            file.seek(-1, 1)
            length = file.read(4)
            length = int.from_bytes(length, 'big')
            dump = file.read(length)
            obj = pickle.loads(dump)
            yield obj

def zip_folder(filename,parser):
    with zipfile.ZipFile(filename, 'r') as zfile:
        for name in zfile.namelist():
            yield KeyValuePair(name, parser(zfile.read(name)))




def folder(location: Union[Path, str], pattern: Optional[str] = None):
    if isinstance(location,str):
        location = Path(location)
    elif isinstance(location,Path):
        pass
    else:
        raise ValueError('Location should be either str or Path, but was {0}, {1}'.format(type(location),location))

    if not os.path.isdir(location):
        raise ValueError('{0} is not a directory'.format(location))

    return location.glob(pattern)






from typing import *
from enum import Enum

T = TypeVar('T')
TOut = TypeVar('TOut')
TKey = TypeVar('TKey')
TValue = TypeVar('TValue')
TFactory = TypeVar('TFactory')

class KeyValuePair(Generic[TKey,TValue]):
    def __init__(self, key: TKey, value: TValue):
        self.key = key
        self.value = value

    def __repr__(self):
        return dict(key=self.key,value=self.value).__repr__()

    def __str__(self):
        return dict(key=self.key, value=self.value).__str__()

class ItemWithIndex(KeyValuePair):
    def __init__(self, key, value):
        super(ItemWithIndex, self).__init__(key,value)
        self.index = key
        self.item = value

class ForkContext:
    def __init__(self, data_to_send):
        self.sent = data_to_send
        self.received = None
        self.result = None

class LoopEndType(Enum):
    NotEqual = 0
    Equal = 1
    Force = 2


from .._misc import *
from collections import OrderedDict


def intersect(en1, en2):
    en2=set(en2)
    for e in en1:
        if e in en2:
            yield e

def concat(en1, en2):
    for e in en1:
        yield e
    for e in en2:
        yield e


def skip_while(en, condition):
    make_yield = False
    for e in en:
        if not make_yield and not condition(e):
            make_yield = True
        if make_yield:
            yield e

def take_while(en, condition):
    for e in en:
        if not condition(e):
            break
        yield e


class Group(KeyValuePair[TKey,List[TValue]]):
    def __init__(self, key: TKey, values: List[TValue]):
        super(Group, self).__init__(key,values)

    def __iter__(self):
        for e in self.value:
            yield e

    def __len__(self):
        return len(self.value)


def group_by(en,selector):
    groups = OrderedDict()
    for e in en:
        key = selector(e)
        if key not in groups:
            groups[key]=[e]
        else:
            groups[key].append(e)

    for key, value in groups.items():
        yield Group(key,value)



def aggregate(en, aggregator):
    value = None
    firstTime = True
    for e in en:
        if firstTime:
            firstTime=False
            value = e
        else:
            value = aggregator(value,e)
    return value


def distinct(en, selector):
    seen = set()
    for e in en:
        key = e
        if selector is not None:
            key = selector(e)
        if key in seen:
            continue
        seen.add(key)
        yield e



from typing import *



from ._queryable_helpers import *
from ._push_queries import *
from . import agg
import itertools


T = TypeVar('T')
TOut = TypeVar('TOut')

class _NoDefault:
    pass

class Queryable(Generic[T]):
    def __init__(self, en: Iterable[T], length: Optional[int] = None):
        self.en = en
        self.length = length

    def __iter__(self):
        for e in self.en:
            yield e

    def _method_concatenator(self, aggregator: AbstractPushQueryElement):
        return aggregator(self.en)

    ## Core methods

    def select(self, selector: Callable[[T], TOut]) -> 'Queryable[TOut]':
        return Queryable(map(selector, self.en), self.length)

    def where(self, filterSelector: Callable[[T], bool]) -> 'Queryable[T]':
        return Queryable(filter(filterSelector, self.en))

    def distinct(self, selector: Optional[Callable[[T], TOut]] = None) -> 'Queryable[T]':
        return Queryable(distinct(self.en, selector))

    def select_many(self, selector: Callable[[T], Iterable[TOut]]) -> 'Queryable[TOut]':
        return Queryable(itertools.chain.from_iterable(map(selector, self.en)))

    def with_indices(self) -> 'Queryable[ItemWithIndex[T]]':
        return Queryable(with_indices(self.en), self.length)

    def group_by(self, selector: Callable[[T], TKey]) -> 'Queryable[Group[TKey,T]]':
        return Queryable(group_by(self.en, selector))


    def skip(self, count) -> 'Queryable[T]':
        return Queryable(itertools.islice(self.en, count, None))

    def take(self, count) -> 'Queryable[T]':
        return Queryable(itertools.islice(self.en, count))

    def skip_while(self, condition) -> 'Queryable[T]':
        return Queryable(skip_while(self.en, condition))

    def take_while(self, condition) -> 'Queryable[T]':
        return Queryable(take_while(self.en, condition))

    def append(self, *args: T) -> 'Queryable[T]':
        return Queryable(append(self.en, args))

    def prepend(self, *args: T) -> 'Queryable[T]':
        return Queryable(prepend(self.en, args))

    def intersect(self, en2: Iterable[T]) -> 'Queryable[T]':
        return Queryable(intersect(self.en, en2))

    def concat(self, en2: Iterable[T]) -> 'Queryable[T]':
        return Queryable(concat(self.en, en2))

    def order_by(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        return Queryable(Orderer(self.en, [(-1, selector)]), self.length)

    def order_by_descending(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        return Queryable(Orderer(self.en, [(1, selector)]), self.length)

    def then_by(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        if not isinstance(self.en, Orderer):
            raise ValueError('then_by can only be called directly after order_by or order_by_descending')
        return Queryable(Orderer(self.en, self.en._funcs + [(-1, selector)]), self.length)

    def then_by_descending(self, selector: Callable[[T], Any]) -> 'Queryable[T]':
        if not isinstance(self.en, Orderer):
            raise ValueError('then_by can only be called directly after order_by or order_by_descending')
        return Queryable(Orderer(self.en, self.en._funcs + [(1, selector)]), self.length)

    def foreach(self, action: Callable[[T], None]) -> None:
        foreach(self.en, action)

    def foreach_and_continue(self, action: Callable[[T], None]) -> 'Queryable[T]':
        return Queryable(foreach_and_continue(self.en, action), self.length)


    def fork(self, context: ForkContext, pipeline: Callable[[Any, Iterable[T]], Any]) -> 'Queryable[T]':
        return Queryable(fork(self.en, context, pipeline))

    def fire_and_forget(self, pipeline: Callable[[Iterable[T]], Any]) -> 'Queryable[T]':
        return Queryable(fork(self.en, ForkContext(None), lambda q, _: pipeline(q)))

    def parallel_select(self, selector, workers_count=None, buffer_size=1) -> 'Queryable[T]':
        return Queryable(parallel_select(self.en, selector, workers_count, buffer_size), self.length)

    def feed(self, *collectors: Callable):
        result = self
        for collector in collectors:
            result = collector(result)
        return result

    def aggregate(self, aggregator):
        return aggregate(self.en, aggregator)

    ## Aggregation methods



    def first(self) -> T:
        return self._method_concatenator(agg.First())

    def first_or_default(self, default=None) -> Optional[T]:
        return self._method_concatenator(agg.First().default_if_empty(default))

    def last(self) -> T:
        return self._method_concatenator(agg.Last())

    def last_or_default(self, default=None) -> Optional[T]:
        return self._method_concatenator(agg.Last().default_if_empty(default))

    def single(self) -> T:
        return self._method_concatenator(agg.Single())

    def single_or_default(self, default=None) -> Optional[T]:
        return self._method_concatenator(agg.Single().default_if_empty(default))

    def min(self, selector : Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> T:
        a = agg.Min(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def max(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> T:
        a = agg.Max(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def argmin(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> TOut:
        a = agg.ArgMin(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def argmax(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> TOut:
        a = agg.ArgMax(selector)
        if not isinstance(default, _NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def sum(self) -> T:
        return self._method_concatenator(agg.Sum())

    def mean(self) -> float:
        return self._method_concatenator(agg.Mean())

    def count(self) -> int:
        return self._method_concatenator(agg.Count())

    def any(self, selector: Optional[Callable] = None) -> bool:
        return self._method_concatenator(agg.Any(selector))

    def all(self, selector: Optional[Callable] = None) -> bool:
        return self._method_concatenator(agg.All(selector))

    def to_list(self) -> List[T]:
        return self._method_concatenator(agg.ToList())

    def to_dictionary(self,
                      key_selector: Optional[Callable[[T], TKey]] = None,
                      value_selector: Optional[Callable[[T], TValue]] = None) -> Dict[TKey, TValue]:
        return self._method_concatenator(agg.ToDictionary(key_selector, value_selector))

    def to_set(self) -> Set[T]:
        return self._method_concatenator(agg.ToSet())

    def to_tuple(self) -> Tuple[T,...]:
        return self._method_concatenator(agg.ToTuple())

    def aggregate_with(self, *args: PushQueryElement):
        if len(args)==1:
            return self._method_concatenator(args[0])
        else:
            return self._method_concatenator(SplitPipelines(*args))


    def to_text_file(self, filename: Union[str, Path], autoflush: bool = False, separator: str = '\n', **kwargs) -> None:
        return self._method_concatenator(agg.ToTextFile(filename, autoflush, separator, **kwargs))

    def to_zip_text_file(self, filename: Union[str, Path], separator: str = '\n', **kwargs) -> None:
        return self._method_concatenator(agg.ToZipTextFile(filename, separator, **kwargs))

    def to_pickle_file(self, filename) -> None:
        return self._method_concatenator(agg.ToPickleFile(filename))

    def to_zip_folder(self, filename: Union[str, Path], writer: Callable = pickle.dumps, replace=True,
                      compression=zipfile.ZIP_DEFLATED) -> None:
        return self._method_concatenator(agg.ToZipFolder(filename, writer, replace, compression))







from multiprocessing import Process, Queue, Pipe, Pool
from .._misc import ForkContext


class _QueueEnd:
    pass


class _QueueAsIterable:
    def __init__(self, queue: Queue):
        self.queue = queue

    def __iter__(self):
        while(True): # actually covered, but not recognized due to being executed in anoher process # pragma: no cover
            e = self.queue.get()
            if isinstance(e,_QueueEnd):
                break
            yield e




def _fork_process(selector, fork_enum, context, pipe): # actually covered, but not recognized due to being executed in anoher process # pragma: no cover
    result = selector(fork_enum,context)
    pipe.send(context)
    pipe.send(result)



def fork(en, context, selector):
    queue = Queue()
    fork_enum = _QueueAsIterable(queue)
    parent_conn, child_conn = Pipe()
    process = Process(target=_fork_process, args=(selector, fork_enum, context.sent, child_conn))
    process.start()
    for e in en:
        queue.put(e)
        yield e
    queue.put(_QueueEnd())
    context.received = parent_conn.recv()
    context.result = parent_conn.recv()
    process.join()


def parallel_select(q,selector, workers_count, buffer_size):
    pool = Pool(workers_count)
    for r in pool.imap(selector,q, buffer_size):
        yield r
    pool.close()


from ._misc import *
from ._push_queries import *
from .agg import *
from . import agg

class _NoDefault:
    pass


class PushQuery(PushQueryElement):
    def __init__(self):
        self.head = None  # type: Optional[PushQueryElement]
        self.tail = None  # type: Optional[PushQueryElement]


    def _method_concatenator(self, pqe: AbstractPushQueryElement):
        if self.head is None:
            self.head = pqe
            self.tail = pqe
        else:
            self.tail.subscribe(pqe, None)
            self.tail = pqe
        return self

    def instance(self):
        if self.head is None:
            raise ValueError('PQEBuilder was not completed, HEAD is not set!')
        return self.head.instance()

    def select(self, selector: Callable) -> 'PushQuery':
        return self._method_concatenator(SelectPQE(selector))

    def where(self, filter: Callable) ->'PushQuery':
        return self._method_concatenator(WherePQE(filter))

    def select_many(self, selector: Callable) -> 'PushQuery':
        return self._method_concatenator(SelectManyPQE(selector))

    def split_pipelines(self, **kwargs:PushQueryElement):
        pqe = SplitPipelines(**kwargs)
        return self._method_concatenator(pqe)

    def split_by_group(self, group_selector, with_total = None):
        return self._method_concatenator(SplitByGroup(group_selector, with_total))

    def split_dictionary(self):
        return self._method_concatenator(SplitByDictionary())





    def first(self) -> 'PushQuery':
        return self._method_concatenator(agg.First())

    def first_or_default(self, default=None) -> 'PushQuery':
        return self._method_concatenator(agg.First().default_if_empty(default))

    def last(self) -> 'PushQuery':
        return self._method_concatenator(agg.Last())

    def last_or_default(self, default=None) -> 'PushQuery':
        return self._method_concatenator(agg.Last().default_if_empty(default))

    def single(self) -> 'PushQuery':
        return self._method_concatenator(agg.Single())

    def single_or_default(self, default=None) -> 'PushQuery':
        return self._method_concatenator(agg.Single().default_if_empty(default))

    def min(self, selector : Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> 'PushQuery':
        a = agg.Min(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def max(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> 'PushQuery':
        a = agg.Max(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def argmin(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> 'PushQuery':
        a = agg.ArgMin(selector)
        if not isinstance(default,_NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def argmax(self, selector: Callable[[T], TOut] = lambda z: z, default = _NoDefault()) -> 'PushQuery':
        a = agg.ArgMax(selector)
        if not isinstance(default, _NoDefault):
            a = a.default_if_empty(default)
        return self._method_concatenator(a)

    def sum(self) -> 'PushQuery':
        return self._method_concatenator(agg.Sum())

    def mean(self) -> 'PushQuery':
        return self._method_concatenator(agg.Mean())

    def count(self) -> 'PushQuery':
        return self._method_concatenator(agg.Count())

    def any(self, selector: Optional[Callable] = None) -> 'PushQuery':
        return self._method_concatenator(Any(selector))

    def all(self, selector: Optional[Callable] = None) -> 'PushQuery':
        return self._method_concatenator(agg.All(selector))

    def to_list(self) -> 'PushQuery':
        return self._method_concatenator(agg.ToList())

    def to_dictionary(self,
                      key_selector: Optional[Callable[[T], TKey]] = None,
                      value_selector: Optional[Callable[[T], TValue]] = None) -> 'PushQuery':
        return self._method_concatenator(agg.ToDictionary(key_selector, value_selector))

    def to_set(self) -> 'PushQuery':
        return self._method_concatenator(agg.ToSet())

    def to_tuple(self) -> 'PushQuery':
        return self._method_concatenator(agg.ToTuple())

    def aggregate_with(self, *args: PushQueryElement):
        if len(args)==1:
            return self._method_concatenator(args[0])
        else:
            return self._method_concatenator(SplitPipelines(*args))





from .._misc import LoopEndType

class loop_maker:
    def __init__(self, begin, delta, end, endType: LoopEndType):
        self.begin = begin
        self.delta = delta
        self.end = end
        self.endType = endType
        self.less_comparison = None

    def _cmp_less_like(self, left, right):
        if self.less_comparison is None: #pragma: no cover
            raise ValueError('loop_maker error: cmp is requested, but less_comparison was not set')
        if self.less_comparison:
            return left<right
        else:
            return left>right

    def make(self):
        value = self.begin

        if self.end is not None:
            if value==self.end:
                if self.endType != LoopEndType.NotEqual:
                    yield value
                return

        yield value

        value = value + self.delta
        if value > self.begin:
            self.less_comparison = True
        else:
            self.less_comparison = False

        if self.end is not None and not self._cmp_less_like(self.begin,self.end):
            raise ValueError('`end` and `delta` parameters are contradictory. ')

        while True:
            if self.end is None:
                yield value
                value+=self.delta
                continue

            if value == self.end:
                if self.endType != LoopEndType.NotEqual:
                    yield value
                break

            elif self._cmp_less_like(value,self.end):
                yield value
                value+=self.delta
                continue

            else:
                if self.endType == LoopEndType.Force:
                    yield self.end
                break


from typing import Union, Callable
from .._misc import KeyValuePair
from .._push_queries.arch import PushQueryElement


from pathlib import Path
import gzip
import pickle
import zipfile
import os


class ToTextFile(PushQueryElement):
    def __init__(self, filename: Union[str, Path], autoflush: bool = False, separator: str = '\n', **kwargs):
        self.filename = filename
        self.autoflush = autoflush
        self.separator = separator
        self.kwargs = kwargs

    def on_enter(factory, instance):
        instance.file = open(factory.filename, 'w', **factory.kwargs)

    def on_process(factory, instance, element):
        instance.file.write(element + factory.separator)
        if factory.autoflush:
            instance.file.flush()

    def on_report(factory, instance):
        return None

    def on_exit(factory, instance, exc_type, exc_val, exc_tb):
        instance.file.__exit__(exc_type, exc_val, exc_tb)


class ToZipTextFile(PushQueryElement):
    def __init__(self, filename: Union[str, Path], separator: str = '\n', **kwargs):
        self.filename = filename
        self.separator = separator
        self.kwargs = kwargs

    def on_enter(factory, instance):
        instance.file = gzip.open(factory.filename, 'wb', **factory.kwargs)

    def on_process(factory, instance, element):
        element += factory.separator
        element = bytes(element, 'utf-8')
        instance.file.write(element)

    def on_report(factory, instance):
        return None

    def on_exit(factory, instance, exc_type, exc_val, exc_tb):
        instance.file.__exit__(exc_type, exc_val, exc_tb)


class ToPickleFile(PushQueryElement):
    def __init__(self, filename: Union[str, Path], file_opener=None):
        self.filename = filename
        if file_opener is None:
            file_opener = lambda fname: open(fname, 'wb')
        self.file_opener = file_opener

    def on_enter(factory, instance):
        instance.file = factory.file_opener(factory.filename)

    def on_process(factory, instance, element):
        dump = pickle.dumps(element)
        length = len(dump)
        length_bytes = length.to_bytes(4, byteorder='big')
        instance.file.write(length_bytes)
        instance.file.write(dump)

    def on_report(factory, instance):
        return None

    def on_exit(factory, instance, exc_type, exc_val, exc_tb):
        instance.file.__exit__(exc_type, exc_val, exc_tb)


class ToZipFolder(PushQueryElement):
    def __init__(self, filename: Union[str, Path], writer: Callable = pickle.dumps, replace=True,
                 compression=zipfile.ZIP_DEFLATED):
        self.filename = filename
        self.compression = compression
        self.writer = writer
        self.replace = replace

    def on_enter(factory, instance):
        if factory.replace and os.path.isfile(factory.filename):
            os.remove(factory.filename)
        instance.file = zipfile.ZipFile(factory.filename, 'a', compression=factory.compression)

    def on_process(factory, instance, element):
        if not isinstance(element, KeyValuePair):
            raise ValueError('Item must be a key-value pair, but was {0}'.format(element))
        instance.file.writestr(element.key, factory.writer(element.value))

    def on_report(factory, instance):
        return None

    def on_exit(factory, instance, exc_type, exc_val, exc_tb):
        instance.file.__exit__(exc_type, exc_val, exc_tb)





'''Classes which support the Queryable interface.'''

# Copyright (c) 2011-2016 Sixty North.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import *
from functools import total_ordering as totally_ordered
import heapq


class Orderer:
    '''A Queryable representing an ordered iterable.

    The sorting implemented by this class is an incremental partial sort so
    you don't pay for sorting results which are never enumerated.'''

    def __init__(self, iterable, orderers: List[Tuple[int,Callable]]):
        '''Create an OrderedIterable.

            Args:
                iterable: The iterable sequence to be ordered.
                order: +1 for ascending, -1 for descending.
                func: The function to select the sorting key.
        '''
        self._iterable = iterable
        self._funcs = orderers


    def __iter__(self):
        '''Support for the iterator protocol.

        Returns:
            An iterator object over the sorted elements.
        '''

        # Determine which sorting algorithms to use
        directions = [direction for direction, _ in self._funcs]
        direction_total = sum(directions)
        if direction_total == -len(self._funcs):
            # Uniform ascending sort - do nothing
            MultiKey = tuple

        elif direction_total == len(self._funcs):
            # Uniform descending sort - invert sense of operators
            @totally_ordered
            class MultiKey(object):
                def __init__(self, t):
                    self.t = tuple(t)

                def __lt__(lhs, rhs):
                    # Uniform descending sort - swap the comparison operators
                    return lhs.t > rhs.t

                def __eq__(lhs, rhs):
                    return lhs.t == rhs.t
        else:
            # Mixed ascending/descending sort - override all operators
            @totally_ordered
            class MultiKey(object):
                def __init__(self, t):
                    self.t = tuple(t)

                # TODO: [asq 1.1] We could use some runtime code generation here to compile a custom comparison operator
                def __lt__(lhs, rhs):
                    for direction, lhs_element, rhs_element in zip(directions, lhs.t, rhs.t):
                        cmp = (lhs_element > rhs_element) - (rhs_element > lhs_element)
                        if cmp == direction:
                            return True
                        if cmp == -direction:
                            return False
                    return False # pragma: no cover

                def __eq__(lhs, rhs):
                    return lhs.t == rhs.t

        # Uniform ascending sort - decorate, sort, undecorate using tuple element
        def create_key(index, item):
            return MultiKey(func(item) for _, func in self._funcs)

        lst = [(create_key(index, item), index, item) for index, item in enumerate(self._iterable)]
        heapq.heapify(lst)
        while lst:
            key, index, item = heapq.heappop(lst)
            yield item



import pickle
import json
import os
from .obj import Obj


class FileIO:

    @staticmethod
    def read_bytes(filename):
        with open(filename, 'rb') as file:
            return file.read()

    @staticmethod
    def read_pickle(filename):
        with open(filename, 'rb') as file:
            return pickle.load(file)

    @staticmethod
    def read_json(filename):
        with open(filename, 'r', encoding='utf-8') as file:
            result = json.load(file)
            return result

    @staticmethod
    def read_text(filename, encoding='utf-8'):
        with open(filename, 'r', encoding=encoding) as file:
            return file.read()

    @staticmethod
    def write_pickle(data, filename):
        with open(filename, 'wb') as file:
            pickle.dump(data, file)

    @staticmethod
    def write_json(data, filename):
        with open(filename, 'w') as file:
            json.dump(data, file, indent=1)

    @staticmethod
    def write_text(data, filename):
        with open(filename, 'w', encoding='utf-8') as file:
            file.write(data)

    @staticmethod
    def write_bytes(data, filename):
        with open(filename, 'wb') as file:
            file.write(data)




from .core import *
from .extended import *
from .original_asq_ordering import *
from .parallel import *
from .loop import *
from .files import *
from .combinatorics import *
from .pandas_df_iter import *


from typing import *
from ._misc import *
from ._queryable_helpers import *
from .queryable import Queryable
from .push_query import PushQuery




class FileQuery:
    def text(self, path: Union[str,Path], **file_kwargs) -> Queryable[str]:
        return Queryable(text_file(path, **file_kwargs))

    def zipped_text(self, path: Union[str,Path], encoding: str ='utf-8') -> Queryable[str]:
        return Queryable(zip_text_file(path, encoding))

    def zipped_folder(self, path: Union[str,Path], parser: Callable = pickle.loads, with_length:bool = True):
        length = None
        if with_length:
            with zipfile.ZipFile(path, 'r') as zfile:
                length = len(zfile.namelist())
        return Queryable(zip_folder(path,parser), length)


    def pickle(self,
               path: Union[str,Path],
               file_obj_factory:Optional[Callable] = lambda fname: open(fname,'rb')
               ) -> Queryable[Any]:
        return Queryable(pickle_file(path, file_obj_factory))


class CombinatoricsQuery:
    def grid(self, **kwargs)->Queryable[Obj]:
        return Queryable(*grid_dict(kwargs))

    def grid_dict(self, dict) -> Queryable[Obj]:
        return Queryable(*grid_dict(dict))

    def cartesian(self,*args)->Queryable[Tuple]:
        return Queryable(*grid_args(args))

    def triangle(self, items: T, with_diagonal=True) -> Queryable[Tuple[T, T]]:
        return Queryable(triangle_iter(items, with_diagonal), (len(items) * (len(items) - 1)) // 2)

    def powerset(self, iterable: Iterable[T]) -> Queryable[Tuple[T,...]]:
        return Queryable(powerset_iter(iterable))

class QueryClass:
    def en(self, en: Iterable[T]) -> Queryable[T]:
        length = None
        if hasattr(en, '__len__'):
            length = len(en)
        return Queryable(en,length)

    def args(self, *args: T)  -> Queryable[T]:
        return Queryable(args,len(args))


    def dict(self, dictionary: Dict[TKey, TValue])  -> Queryable[KeyValuePair[TKey, TValue]]:
        return Queryable(dictionary.items(),len(dictionary)).select(lambda z: KeyValuePair(z[0],z[1]))

    def push(self) -> PushQuery:
        return PushQuery()

    def loop(self, begin: T, delta: T, end: Optional[T] = None, endtype=LoopEndType.NotEqual) -> Queryable[T]:
        lp = loop_maker(begin,delta,end,endtype)
        return Queryable(lp.make())

    file = FileQuery()

    combinatorics = CombinatoricsQuery()

    def folder(self, location: Union[Path, str], pattern: str = '*') -> Queryable[Path]:
        return Queryable(folder(location, pattern))



Query = QueryClass()


from .._misc import *

def with_indices(en):
    for index,item in enumerate(en):
        yield ItemWithIndex(index,item)


def foreach(en, action):
    for element in en:
        if action is not None:
            action(element)

def foreach_and_continue(en,action):
    for element in en:
        action(element)
        yield element

def append(en,array):
    for e in en:
        yield e
    for e in array:
        yield e

def prepend(en, array):
    for e in array:
        yield e
    for e in en:
        yield e


